@@ -3,21 +3,25 @@ require 'sass/tree/node'
 module Sass::Tree
   class AttrNode < ValueNode
     attr_accessor :name
-    
+
     def initialize(name, value, style)
       @name = name
       super(value, style)
     end
-    
+
+    def ==(other)
+      self.class == other.class && name == other.name && super
+    end
+
     def to_s(tabs, parent_name = nil)
       if value[-1] == ?;
-        raise Sass::SyntaxError.new("Invalid attribute: #{declaration.dump} (This isn't CSS!)", @line)
+        raise Sass::SyntaxError.new("Invalid attribute: #{declaration.dump} (This isn't CSS!).", @line)
       end
       real_name = name
       real_name = "#{parent_name}-#{real_name}" if parent_name
 
       if value.empty? && children.empty?
-        raise Sass::SyntaxError.new("Invalid attribute: #{declaration.dump}", @line)
+        raise Sass::SyntaxError.new("Invalid attribute: #{declaration.dump}.", @line)
       end
 
       join_string = case @style
@@ -34,7 +38,7 @@ module Sass::Tree
       children.each do |kid|
         to_return << "#{kid.to_s(tabs, real_name)}" << join_string
       end
-            
+
       (@style == :compressed && parent_name) ? to_return : to_return[0...-1]
     end
 
